import json
import os
from dataclasses import dataclass
from typing import Any, Callable, Iterator

import requests

from ..driver import OS
from ..utils import Filters, Some, async_cached, choose, run_ok, run_output, unsome
from .archive import ArchivePackage


@async_cached
async def have_github_auth() -> bool:
    return await run_ok("gh", "auth", "status")


async def github_api(url: str) -> Any:
    if await have_github_auth():
        return json.loads(await run_output("gh", "api", url))
    else:
        try:
            token = os.environ["GITHUB_TOKEN"]
        except KeyError:
            token = None

        headers = {}
        if token:
            headers["Authorization"] = f"token {token}"

        result = requests.get(f"https://api.github.com/{url}", headers=headers)
        result.raise_for_status()

        return result.json()


@dataclass
class GitHubReleaseArtifact:
    name: str
    url: str


@dataclass
class GitHubRelease:
    tag_name: str
    assets: list[GitHubReleaseArtifact]


class GitHubPackage(ArchivePackage):
    prefixes: list[str]
    suffixes: list[str]
    excludes: list[str]

    def __init__(
        self,
        *,
        repo: str,
        prefix: Some[str] = None,
        suffix: Some[str] = None,
        exclude: Some[str] = None,
        regex: Some[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.repo = repo
        self.prefixes = unsome(prefix)
        self.suffixes = unsome(suffix)
        self.excludes = unsome(exclude)
        self.regex = unsome(regex)

    @async_cached
    async def latest_release(self) -> GitHubRelease:
        latest = await github_api(f"repos/{self.repo}/releases/latest")
        return GitHubRelease(
            tag_name=latest["tag_name"],
            assets=[
                GitHubReleaseArtifact(
                    name=result["name"], url=result["browser_download_url"]
                )
                for result in latest["assets"]
            ],
        )

    def filters(self, target_os: OS) -> Iterator[Callable[[str], bool]]:
        for prefix in self.prefixes:
            yield Filters.startswith(prefix)
        for suffix in self.suffixes:
            yield Filters.endswith(suffix)
        for exclude in self.excludes:
            yield Filters.excludes(exclude)
        for regex in self.regex:
            yield Filters.regex(regex)
        for hint in [".tar.gz"]:
            yield Filters.includes(hint)
        for signature_hint in [".asc", ".sig", "sha256"]:
            yield Filters.excludes(signature_hint)
        for os_hint in target_os.switch(
            linux=[
                Filters.includes("linux"),
                Filters.includes("gnu"),
                Filters.excludes("musl"),
            ],
            macos=[Filters.includes(hint) for hint in ["macos", "darwin", "osx"]],
        ):
            yield os_hint
        arch_hints = ["x86_64", "amd64"]
        for hint in arch_hints:
            yield Filters.includes(hint)

    async def artifact(self) -> GitHubReleaseArtifact:
        candidates = (await self.latest_release()).assets

        def candidate_filter(
            name_filter: Callable[[str], bool]
        ) -> Callable[[GitHubReleaseArtifact], bool]:
            return lambda candidate: name_filter(candidate.name)

        target_os = await self.driver.os()
        return choose(candidates, map(candidate_filter, self.filters(target_os)))

    async def archive_url(self) -> str:
        return (await self.artifact()).url

    @property
    def name(self) -> str:
        return self.repo

    async def get_remote_version(self) -> str:
        return (await self.latest_release()).tag_name



from pathlib import Path
from typing import Any, Tuple, Type

import yaml

from ..driver import Driver
from ..state import DB
from .base import Package
from .clone import Clone
from .github import GitHubPackage
from .links import Links
from .pip import PipPackage
from .pipx import PipxPackage
from .shell import Shell
from .system import SystemPackage
from .url import URLPackage
from .yum_repo import YumRepo

TYPES: list[Tuple[str, Type[Package]]] = [
    ("name", SystemPackage),
    ("repo", GitHubPackage),
    ("url", URLPackage),
    ("clone", Clone),
    ("pip", PipPackage),
    ("pipx", PipxPackage),
    ("shell", Shell),
    ("links", Links),
    ("yum_name", YumRepo),
]


def parse_package(package: Any, *, db: DB, driver: Driver) -> Package:
    if not isinstance(package, dict):
        raise ValueError(f"Dictionary expected, got: {package}.")

    for key, package_type in TYPES:
        if key in package:
            return package_type(**package, db=db, driver=driver)

    keys = ", ".join(f"'{key}'" for key, _ in TYPES)
    raise ValueError(f"Either of {keys} must be present, got: {package}.")


def load_packages(component: str, *, db: DB, driver: Driver) -> list[Package]:
    with open(Path("packages") / f"{component}.yaml") as f:
        packages = yaml.safe_load(f)
        return [parse_package(package, db=db, driver=driver) for package in packages]



import dataclasses
import sqlite3
import threading
from abc import ABCMeta, abstractmethod
from typing import Callable, Generic, Type, TypeVar

T = TypeVar("T")

DB_PATH = "state.sqlite"


class DB:
    """Thread-safe SQLite database connection."""

    def __init__(self, path: str) -> None:
        self.path = path
        self.cache = threading.local()

    @property
    def instance(self) -> sqlite3.Connection:
        try:
            return self.cache.connection
        except AttributeError:
            pass

        connection = sqlite3.connect(self.path)
        connection.row_factory = sqlite3.Row
        self.cache.connection = connection
        return connection


class Storage(Generic[T], metaclass=ABCMeta):
    @abstractmethod
    def __getitem__(self, key: str) -> T:
        pass

    @abstractmethod
    def __delitem__(self, key: str) -> None:
        pass

    @abstractmethod
    def __setitem__(self, key: str, value: T) -> None:
        pass


StorageDefinition = Callable[[DB], Storage[T]]


def storage(name: str, klass: Type[T]) -> StorageDefinition[T]:
    attributes: list[str] = [field.name for field in dataclasses.fields(klass)]

    def storage_table(db: DB) -> Storage[T]:
        db.instance.execute(
            f'CREATE TABLE IF NOT EXISTS {name} (id TEXT PRIMARY KEY, {", ".join(attributes)})'
        )

        class StorageImpl(Storage[T]):
            def __getitem__(self, key: str) -> T:
                row = db.instance.execute(
                    f"SELECT * FROM {name} WHERE id = ?", (key,)
                ).fetchone()
                if row:
                    attributes = {key: row[key] for key in row.keys() if key != "id"}
                    return klass(**attributes)
                raise KeyError(key)

            def __delitem__(self, key: str) -> None:
                db.instance.execute(f"DELETE FROM {name} WHERE id = ?", (key,))

            def __setitem__(self, key: str, value: T) -> None:
                with db.instance:
                    del self[key]
                    attr_clause = "?"
                    attr_values = [key]
                    for attribute in attributes:
                        attr_clause += ", ?"
                        attr_values.append(getattr(value, attribute))
                    db.instance.execute(
                        f"INSERT INTO {name} VALUES ({attr_clause})", attr_values
                    )

        return StorageImpl()

    return storage_table



import re
import subprocess
from functools import wraps
from pathlib import Path
from typing import (
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Iterable,
    Iterator,
    Optional,
    TypeVar,
    Union,
    overload,
)

import requests
import tqdm  # type: ignore
import trio

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")


Some = Optional[Union[T, list[T]]]


def unsome_(x: Some[T]) -> Optional[list[T]]:
    if x is None:
        return None
    if isinstance(x, list):
        return x
    return [x]


def unsome(x: Some[T]) -> list[T]:
    return unsome_(x) or []


TERMINAL_LOCK = trio.Lock()


async def run(*args: Union[str, Path]) -> subprocess.CompletedProcess:
    return await trio.run_process(args, check=True)


async def run_ok(*args: Union[str, Path]) -> bool:
    try:
        return (
            await trio.run_process(
                args,
                check=False,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
        ).returncode == 0
    except FileNotFoundError:
        return False


async def run_output(*args: Union[str, Path]) -> str:
    return (
        (await trio.run_process(args, capture_stdout=True, check=True))
        .stdout.decode()
        .strip()
    )


def flatten(items: Iterable[Iterable[T]]) -> list[T]:
    return [item for sublist in items for item in sublist]


async def parallel_map_tqdm(
    action: Callable[[T], Awaitable[U]], items: list[T]
) -> list[U]:
    results: list[U] = []

    with tqdm.tqdm(total=len(items)) as progress:
        async with trio.open_nursery() as nursery:

            async def action_and_update(item: T) -> None:
                result = await action(item)
                async with TERMINAL_LOCK:
                    progress.update(1)
                results.append(result)

            for item in items:
                nursery.start_soon(action_and_update, item)

            return results


async def gather(*tasks: Callable[[], Awaitable[T]]) -> list[T]:
    async def collect(
        index: int, task: Callable[[], Awaitable[T]], results: dict[int, T]
    ):
        results[index] = await task()

    results: dict[int, T] = {}
    async with trio.open_nursery() as nursery:
        for index, task in enumerate(tasks):
            nursery.start_soon(collect, index, task, results)
    return [results[i] for i in range(len(tasks))]


class Filters:
    @staticmethod
    def includes(substring: str) -> Callable[[str], bool]:
        return lambda x: substring in x.lower()

    @staticmethod
    def excludes(substring: str) -> Callable[[str], bool]:
        return lambda x: substring not in x

    @staticmethod
    def startswith(prefix: str) -> Callable[[str], bool]:
        return lambda x: x.startswith(prefix)

    @staticmethod
    def endswith(suffix: str) -> Callable[[str], bool]:
        return lambda x: x.endswith(suffix)

    @staticmethod
    def regex(regex: str) -> Callable[[str], bool]:
        regex_compiled = re.compile(regex)

        return lambda x: regex_compiled.match(x) is not None


def choose(candidates: list[T], filters: Iterator[Callable[[T], bool]]) -> T:
    if len(candidates) == 0:
        raise ValueError("No candidates to choose from.")
    while len(candidates) > 1:
        try:
            filter_fn = next(filters)
        except StopIteration:
            break

        new_candidates = list(filter(filter_fn, candidates))
        if new_candidates:
            candidates = new_candidates

    if len(candidates) == 1:
        return candidates[0]

    raise ValueError(f"Cannot choose between: {candidates}.")


def url_version(url: str) -> str:
    head_response = requests.head(url, allow_redirects=True)
    return head_response.headers["etag"]


def raise_(exception: BaseException) -> Any:
    raise exception


@overload
def async_cached(
    fn: Callable[[], Coroutine[Any, Any, T]]
) -> Callable[[], Coroutine[Any, Any, T]]:
    ...


@overload
def async_cached(
    fn: Callable[[U], Coroutine[Any, Any, T]]
) -> Callable[[U], Coroutine[Any, Any, T]]:
    ...


@overload
def async_cached(
    fn: Callable[[U, V], Coroutine[Any, Any, T]]
) -> Callable[[U, V], Coroutine[Any, Any, T]]:
    ...


def async_cached(
    fn: Callable[..., Coroutine[Any, Any, T]]
) -> Callable[..., Coroutine[Any, Any, T]]:
    return _async_cached_lock(None, fn)


@overload
def async_cached_lock(
    fn: Callable[[], Coroutine[Any, Any, T]]
) -> Callable[[], Coroutine[Any, Any, T]]:
    ...


@overload
def async_cached_lock(
    fn: Callable[[U], Coroutine[Any, Any, T]]
) -> Callable[[U], Coroutine[Any, Any, T]]:
    ...


@overload
def async_cached_lock(
    fn: Callable[[U, V], Coroutine[Any, Any, T]]
) -> Callable[[U, V], Coroutine[Any, Any, T]]:
    ...


def async_cached_lock(
    fn: Callable[..., Coroutine[Any, Any, T]]
) -> Callable[..., Coroutine[Any, Any, T]]:
    return _async_cached_lock(trio.Lock(), fn)


class NoLock:
    async def __aenter__(self) -> None:
        pass

    async def __aexit__(self, *args: Any) -> None:
        pass


NO_LOCK = NoLock()


def _async_cached_lock(
    lock: Optional[trio.Lock], fn: Callable[..., Coroutine[Any, Any, T]]
) -> Callable[..., Coroutine[Any, Any, T]]:
    cache: dict[Any, T] = {}

    @wraps(fn)
    async def cached_fn(*args: Any) -> T:
        if args not in cache:
            async with lock or NO_LOCK:
                # Check again, someone holding the lock might have updated the cache
                if args not in cache:
                    cache[args] = await fn(*args)
        return cache[args]

    return cached_fn



import configparser
import hashlib
from io import StringIO
from pathlib import Path

from ..utils import Optional, raise_
from .manual_version import ManualVersion


class YumRepo(ManualVersion):
    def __init__(
        self, yum_name: str, yum_url: str, gpg_key: Optional[str] = None, **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.repo_name = yum_name
        self.baseurl = yum_url
        self.gpg_key = gpg_key

    @property
    def name(self) -> str:
        return f"yum-{self.repo_name}"

    async def get_remote_version(self) -> str:
        m = hashlib.sha256()
        m.update(self.baseurl.encode())
        m.update((self.gpg_key or "").encode())
        return m.hexdigest()

    async def install(self) -> None:
        (await self.driver.os()).switch_(
            linux=lambda linux: lambda: None
            if linux.distribution == "fedora"
            else raise_(ValueError("YumRepo is only supported on Fedora")),
            macos=lambda: raise_(ValueError("YumRepo is only supported on Linux")),
        )()

        repo = configparser.ConfigParser()
        repo[self.repo_name] = {
            "name": self.repo_name,
            "baseurl": self.baseurl,
            "enabled": "1",
        }
        if self.gpg_key is not None:
            repo[self.repo_name]["gpgcheck"] = "1"
            repo[self.repo_name]["gpgkey"] = self.gpg_key

        contents = StringIO()
        repo.write(contents, space_around_delimiters=False)
        await self.driver.with_root(True).write_file(
            Path(f"/etc/yum.repos.d/{self.repo_name}.repo"), contents.getvalue()
        )

        if self.gpg_key:
            await self.driver.with_root(True).run("rpm", "--import", self.gpg_key)

        await super().install()



from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Iterable, Union

from ..utils import async_cached
from .manual import ManualPackage


class ArchivePackage(ManualPackage, metaclass=ABCMeta):
    def __init__(
        self,
        *,
        raw: Union[bool, str] = False,
        raw_executable: bool = False,
        strip: int = 0,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.raw = raw
        self.raw_executable = raw_executable
        self.strip = strip

    @abstractmethod
    async def archive_url(self) -> str:
        pass

    async def package_directory(self) -> Path:
        result = (await self.local()) / "mybox" / self.name.replace("/", "--")
        await self.driver.makedirs(result)
        return result

    @async_cached
    async def tar(self) -> str:
        return await self.driver.find_executable("gtar", "tar")

    async def untar(self, source: Path, *extra: str) -> None:
        await self.driver.run(
            await self.tar(),
            "-x",
            "--strip",
            str(self.strip),
            "-C",
            await self.package_directory(),
            *extra,
            "-f",
            source,
        )

    async def unzip(self, source: Path) -> None:
        if self.strip > 0:
            raise NotImplementedError("Strip is not supported for unzip.")
        await self.driver.run(
            "unzip", "-o", "-qq", source, "-d", await self.package_directory()
        )

    async def extract(self, url: str, source: Path) -> None:
        if self.raw:
            if isinstance(self.raw, str):
                filename = self.raw
            else:
                filename = url.rsplit("/", 1)[-1]
            target = await self.package_directory() / filename
            await self.driver.run("cp", source, target)
            if self.raw_executable:
                await self.driver.make_executable(target)
        elif url.endswith(".tar"):
            await self.untar(source)
        elif url.endswith(".tar.gz") or url.endswith(".tgz"):
            await self.untar(source, "-z")
        elif url.endswith(".tar.bz2"):
            await self.untar(source, "-j")
        elif url.endswith(".tar.xz") or url.endswith(".txz"):
            await self.untar(source, "-J")
        elif url.endswith(".zip"):
            await self.unzip(source)
        else:
            raise ValueError(f"Unknown archive format: {url}")

    async def binary_path(self, binary: str) -> Path:
        paths: list[list[str]] = [[], ["bin"]]
        for relative_path in paths:
            candidate = await self.package_directory() / Path(*relative_path) / binary
            if await self.driver.is_executable(candidate):
                return candidate
        raise ValueError(f"Cannot find {binary} in {await self.package_directory()}.")

    async def app_path(self, name: str) -> Path:
        candidate = (
            await self.package_directory()
            / "share"
            / "applications"
            / f"{name}.desktop"
        )
        if await self.driver.is_file(candidate):
            return candidate
        raise ValueError(
            f"Cannot find application '{name}' in {await self.package_directory()}."
        )

    async def icon_paths(self, name: str) -> Iterable[Path]:
        return [
            Path(path)
            for path in (
                await self.driver.run_output(
                    "find",
                    await self.package_directory(),
                    "-name",
                    f"{name}.svg",
                    "-o",
                    "-name",
                    f"{name}.png",
                )
            ).splitlines()
        ]

    async def font_path(self, name: str) -> Path:
        candidate = await self.package_directory() / name
        if await self.driver.is_file(candidate):
            return candidate
        raise ValueError(
            f"Cannot find font '{name}' in {await self.package_directory()}."
        )

    async def install(self):
        url = await self.archive_url()
        async with self.driver.tempfile() as archive_path:
            await self.driver.run(
                "curl",
                "--silent",
                "--show-error",
                "--location",
                "--output",
                archive_path,
                url,
            )
            await self.extract(url, archive_path)
        await super().install()



import json

from .pip_base import PipBasePackage


class PipPackage(PipBasePackage):
    def __init__(self, *, pip: str, **kwargs) -> None:
        self.package = pip.lower()
        super().__init__(**kwargs)

    async def get_all_versions(self) -> dict[str, str]:
        packages_json = await self.driver.run_output(
            *self.cmd("list", "--format", "json"), silent=True
        )

        # https://github.com/pypa/pip/issues/11282
        packages_json = packages_json.split("\n\n[notice]", maxsplit=1)[0]

        packages = json.loads(packages_json)

        return {package["name"].lower(): package["version"] for package in packages}

    def cmd(self, cmd: str, /, *args: str) -> list[str]:
        return ["python3", "-m", "pip", cmd, "--user", *args]



from abc import ABCMeta, abstractmethod
from typing import Optional

from ..driver import Driver
from ..state import DB
from ..utils import Some, async_cached, gather, unsome_


class Package(metaclass=ABCMeta):
    os: Optional[list[str]]
    distribution: Optional[list[str]]

    def __init__(
        self,
        *,
        db: DB,
        driver: Driver,
        os: Some[str] = None,
        distribution: Some[str] = None,
    ) -> None:
        self.db = db
        self.driver_ = driver
        self.os = unsome_(os)
        self.distribution = unsome_(distribution)

    @property
    def driver(self) -> Driver:
        return self.driver_

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    async def get_remote_version(self) -> str:
        pass

    @async_cached
    async def remote_version(self) -> str:
        return await self.get_remote_version()

    @abstractmethod
    async def local_version(self) -> Optional[str]:
        pass

    async def is_installed(self) -> bool:
        remote, local = await gather(self.remote_version, self.local_version)
        return remote == local

    @abstractmethod
    async def install(self) -> None:
        pass

    async def applicable(self) -> bool:
        def check_os(name: str) -> bool:
            return self.os is None or name in self.os

        os = await self.driver.os()
        return os.switch_(
            linux=lambda linux: check_os("linux")
            and (self.distribution is None or linux.distribution in self.distribution),
            macos=check_os("darwin"),
        )

    async def ensure(self) -> bool:
        if not await self.applicable():
            return False
        if await self.is_installed():
            return False
        await self.install()
        return True



import argparse
from typing import Optional

import trio
from typed_argparse import TypedArgs

from .driver import LocalDriver
from .package import Package, load_packages
from .state import DB, DB_PATH
from .utils import flatten, parallel_map_tqdm


class Args(TypedArgs):
    component: list[str]


def parse_args(args: Optional[list[str]] = None) -> Args:
    result = argparse.ArgumentParser(description="Install the environment")
    result.add_argument("component", nargs="*", help="The component to install")
    return Args.from_argparse(result.parse_args(args))


async def main():
    db = DB(DB_PATH)
    driver = LocalDriver()
    args = parse_args()
    components: frozenset[str] = frozenset(args.component) | {"base"}
    packages = flatten(
        load_packages(component, db=db, driver=driver) for component in components
    )

    async def process(package: Package) -> bool:
        return await package.ensure()

    async def process_and_record(package: Package) -> Optional[str]:
        if await process(package):
            return package.name
        return None

    results = await parallel_map_tqdm(process_and_record, packages)
    installed = list(filter(None, results))
    if installed:
        print(f"{len(installed)} packages installed or updated: {', '.join(installed)}")
    else:
        print("Everything up to date.")


def sync_main():
    trio.run(main)



from abc import ABCMeta, abstractmethod
from typing import Optional

import requests

from .base import Package


class PipBasePackage(Package, metaclass=ABCMeta):
    package: str

    @property
    def name(self) -> str:
        return self.package

    @abstractmethod
    async def get_all_versions(self) -> dict[str, str]:
        pass

    async def local_version(self) -> Optional[str]:
        return (await self.get_all_versions()).get(self.package)

    async def get_remote_version(self) -> str:
        pypi_info = requests.get(f"https://pypi.org/pypi/{self.package}/json").json()
        return pypi_info["info"]["version"]

    @abstractmethod
    def cmd(self, cmd: str, /, *args: str) -> list[str]:
        pass

    async def install(self) -> None:
        cmd = "install" if await self.local_version() is None else "upgrade"
        await self.driver.run(*self.cmd(cmd, self.package))



import configparser
import re
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Iterable, Optional

from ..utils import Some, async_cached, unsome
from .manual_version import ManualVersion
from .root import Root


class ManualPackage(Root, ManualVersion, metaclass=ABCMeta):
    binaries: list[str]
    apps: list[str]
    fonts: list[str]

    def __init__(
        self,
        *,
        binary: Some[str] = None,
        binary_wrapper: bool = False,
        app: Some[str] = None,
        font: Some[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.binaries = unsome(binary)
        self.binary_wrapper = binary_wrapper
        self.apps = unsome(app)
        self.fonts = unsome(font)

    @async_cached
    async def local(self) -> Path:
        if self.root:
            return Path("/usr/local")
        else:
            return await self.driver.local()

    @abstractmethod
    async def binary_path(self, binary: str) -> Path:
        pass

    async def install_binary(self, name: str) -> None:
        await self.driver.link(
            await self.binary_path(name),
            await self.local() / "bin" / name,
            method="binary_wrapper" if self.binary_wrapper else None,
        )

    @abstractmethod
    async def app_path(self, name: str) -> Path:
        pass

    @abstractmethod
    async def icon_paths(self, name: str) -> Iterable[Path]:
        pass

    RESOLUTION_RE = re.compile(r"(\d+)x(\d+)")

    async def icon_target_path(self, path: Path) -> Path:
        resolution: str
        if path.suffix == ".svg":
            resolution = "scalable"
        elif path.suffix == ".png":
            for part in path.parts:
                if self.RESOLUTION_RE.match(part):
                    resolution = part
                    break
            else:
                # Give a low resolution to prefer ones given explicitly
                resolution = "16x16"
        else:
            raise ValueError(f"Unknown icon type: '{path}'")

        return (
            (await self.local())
            / "share"
            / "icons"
            / "hicolor"
            / resolution
            / "apps"
            / path.name
        )

    async def icon_name(self, app_path: Path) -> Optional[str]:
        config = await self.driver.read_file(app_path)
        app = configparser.ConfigParser()
        app.read_string(config)
        return app["Desktop Entry"].get("Icon")

    async def install_app(self, name: str) -> None:
        await (await self.driver.os()).switch(
            linux=self.install_app_linux, macos=self.install_app_macos
        )(name)

    async def install_app_linux(self, name: str) -> None:
        path = await self.app_path(name)
        target = await self.local() / "share" / "applications" / f"{name}.desktop"
        await self.driver.link(path, target)

        icon = await self.icon_name(path)
        if icon:
            for icon_path in await self.icon_paths(name):
                target = await self.icon_target_path(icon_path)
                await self.driver.link(icon_path, target)

    async def install_app_macos(self, name: str) -> None:
        # FIXME: copy to /Applications and/or ~/Applications; ensure names,
        # etc. are correct
        pass

    @abstractmethod
    async def font_path(self, name: str) -> Path:
        pass

    async def install_font(self, name: str) -> None:
        font_dir = (await self.driver.os()).switch(
            linux=await self.local() / "share" / "fonts",
            macos=await self.driver.home() / "Library" / "Fonts",
        )
        await self.driver.makedirs(font_dir)
        source = await self.font_path(name)
        target = font_dir / name
        await self.driver.link(source, target)
        if await self.driver.executable_exists("fc-cache"):
            await self.driver.run("fc-cache", "-f", font_dir)

    async def install(self) -> None:
        for binary in self.binaries:
            await self.install_binary(binary)
        for app in self.apps:
            await self.install_app(app)
        for font in self.fonts:
            await self.install_font(font)
        await super().install()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mybox', 'mybox.package', 'mybox.state']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'requests>=2.28.1,<3.0.0',
 'tqdm>=4.64.0,<5.0.0',
 'trio>=0.22.0,<0.23.0',
 'typed-argparse>=0.2.4,<0.3.0']

entry_points = \
{'console_scripts': ['mybox = mybox.main:sync_main']}

setup_kwargs = {
    'name': 'mybox',
    'version': '0.2.5',
    'description': 'Manage the configuration and tools on your workstation without bothering the OS too much',
    'long_description': '# Mybox\n\n🖥️ This is a box. 📦 And it is mine. 🐱\n\nThere are many 🍱 nice things in there. I wouldn\'t want 🧰 to be without them.\n\nEven if I move 🏠 or work 🏢 I want to be comfortable.\n\n---\n\nManage the configuration and tools on your workstation without bothering the OS\ntoo much (maybe your favorite one isn\'t supported by `$WORK` or you have\ndifferent ones for different roles).\n\n## Usage\n\n* Run the [bootstrap](bootstrap) script:\n\n  ```shell\n  /bin/bash -c "$(curl -fsSL https://raw.githubusercontent.com/koterpillar/mybox/main/bootstrap)"\n  ```\n\n* Run `mybox` from the directory with package definitions.\n\n  For package definition examples, see\n  [koterpillar/desktop](https://github.com/koterpillar/desktop/).\n\n* To install optional components, add their names as arguments, e.g.\n  `mybox development`.\n\n## Development\n\nPre-requisites (see [install-dev](install-dev) for ways to install):\n\n* [Poetry](https://python-poetry.org/)\n* [ShellCheck](https://www.shellcheck.net/)\n\nRun [`./lint`](lint) to check style & types, `./lint --format` to apply\nformatting automatically.\n\nRun `./test` to execute the project\'s tests.\n\n### Running locally\n\n* Run `poetry install`.\n* Run `poetry shell`.\n* In the launched shell, go to the directory with package definitions.\n* Run `mybox` with the desired arguments.\n\n### Releasing\n\nCreate and push a signed Git tag of the format `vX.Y.Z`. The release will be\ndone using GitHub actions.\n',
    'author': 'Alexey Kotlyarov',
    'author_email': 'a@koterpillar.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/koterpillar/mybox',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from abc import ABCMeta
from functools import cached_property

from ..driver import Driver
from .base import Package


class Root(Package, metaclass=ABCMeta):
    root: bool

    def __init__(self, root: bool = False, **kwargs) -> None:
        super().__init__(**kwargs)
        self.root = root

    @cached_property
    def driver(self) -> Driver:
        return super().driver.with_root(self.root)



import json

from .pip_base import PipBasePackage


class PipxPackage(PipBasePackage):
    def __init__(self, *, pipx: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.package = pipx

    async def get_all_versions(self) -> dict[str, str]:
        pipx_list = json.loads(
            await self.driver.run_output(*self.cmd("list", "--json"), silent=True)
        )
        packages = (
            item["metadata"]["main_package"] for item in pipx_list["venvs"].values()
        )
        return {package["package"]: package["package_version"] for package in packages}

    def cmd(self, cmd: str, /, *args: str) -> list[str]:
        return ["python3", "-m", "pipx", cmd, *args]
